package ru.dbotthepony.kommons.arrays

import java.lang.StringBuilder
import java.lang.ref.Reference
import java.lang.ref.SoftReference
import java.lang.ref.WeakReference
import java.nio.Buffer

internal class MatrixCache<T : Array2D> : ThreadLocal<ArrayList<Reference<T>?>>() {
	override fun initialValue(): ArrayList<Reference<T>?> {
		return ArrayList()
	}
}

// template
private inline fun <T : Array2D> ArrayList<Reference<T>?>.load(index: Int, factory: (Int, Int) -> T): T {
	while (size <= index) add(null)
	var cached = this[index]?.get()

	if (cached == null) {
		cached = factory(index, index)
		this[index] = if (index < 10) SoftReference(cached) else WeakReference(cached)
	}

	return cached
}

// template
private inline fun <N : Number, A : Array2D, B : A> loadMatrixMinor(
	input: A,
	output: B,
	get: (A, Int, Int) -> N,
	set: (B, Int, Int, N) -> Unit,
	column: Int, row: Int
) {
	var tc = 0
	var tr = 0

	for (c in 0 until column) {
		for (r in 0 until row) {
			set.invoke(output, tc, tr++, get.invoke(input, c, r))
		}

		for (r in row + 1 until input.rows) {
			set.invoke(output, tc, tr++, get.invoke(input, c, r))
		}

		tc++
	}

	for (c in column + 1 until input.columns) {
		for (r in 0 until row) {
			set.invoke(output, tc, tr++, get.invoke(input, c, r))
		}

		for (r in row + 1 until input.rows) {
			set.invoke(output, tc, tr++, get.invoke(input, c, r))
		}

		tc++
	}
}

// template
private inline fun <N : Number, Immutable : Array2D, Mutable : Immutable> matrixCofactor(
	input: Immutable,
	output: Mutable,
	adjugate: Boolean,
	get: (Immutable, Int, Int) -> N,
	set: (Mutable, Int, Int, N) -> Unit,
	cache: ArrayList<Reference<Mutable>?>,
	factory: (Int, Int) -> Mutable,
	determinant: (Immutable, ArrayList<Reference<Mutable>?>) -> N,
	neg: (N) -> N
) {
	input.requireIsSquare()
	output.requireIsSquare()
	input.requireSizeEquals(output)

	if (input.columns == 0) {
		return
	} else if (input.columns == 1) {
		set(output, 0, 0, get(input, 0, 0))
		return
	}

	val target = output
	val cached = cache.load(input.columns - 1, factory)

	if (adjugate) {
		for (column in input.columnIndices) {
			for (row in input.rowIndices) {
				loadMatrixMinor(input, cached, get, set, column, row)
				val det = determinant(cached, cache)

				if ((column + row) and 1 == 1) {
					set(target, row, column, neg(det))
				} else {
					set(target, row, column, det)
				}
			}
		}
	} else {
		for (column in input.columnIndices) {
			for (row in input.rowIndices) {
				loadMatrixMinor(input, cached, get, set, column, row)
				val det = determinant(cached, cache)

				if ((column + row) and 1 == 1) {
					set(target, column, row, neg(det))
				} else {
					set(target, column, row, det)
				}
			}
		}
	}
}

// concrete
private val floatCache = MatrixCache<Float2DArray>()
private val doubleCache = MatrixCache<Double2DArray>()
private val intCache = MatrixCache<Int2DArray>()
private val longCache = MatrixCache<Long2DArray>()

// templated
internal fun matrixCofactor(
	input: Float2DArray,
	output: Float2DArray,
	adjugate: Boolean,
) = matrixCofactor(
	input, output, adjugate,
	Float2DArray::get, Float2DArray::set,
	floatCache.get(), Float2DArray::allocate,
	::matrixDeterminant, Float::unaryMinus
)

// templated
internal fun matrixCofactor(
	input: Double2DArray,
	output: Double2DArray,
	adjugate: Boolean,
) = matrixCofactor(
	input, output, adjugate,
	Double2DArray::get, Double2DArray::set,
	doubleCache.get(), Double2DArray::allocate,
	::matrixDeterminant, Double::unaryMinus
)

internal fun matrixCofactor(
	input: Int2DArray,
	output: Int2DArray,
	adjugate: Boolean,
) = matrixCofactor(
	input, output, adjugate,
	Int2DArray::get, Int2DArray::set,
	intCache.get(), Int2DArray::allocate,
	::matrixDeterminant, Int::unaryMinus
)

internal fun matrixCofactor(
	input: Long2DArray,
	output: Long2DArray,
	adjugate: Boolean,
) = matrixCofactor(
	input, output, adjugate,
	Long2DArray::get, Long2DArray::set,
	longCache.get(), Long2DArray::allocate,
	::matrixDeterminant, Long::unaryMinus
)

internal fun matrixDeterminant(input: Float2DArray, cache: ArrayList<Reference<Float2DArray>?> = floatCache.get()): Float {
	input.requireIsSquare()
	require(input.columns >= 1) { "Can't compute determinant of matrix with no elements" }

	if (input.columns == 1) {
		return input[0, 0]
	} else if (input.columns == 2) {
		val a = input[0, 0]
		val b = input[0, 1]
		val c = input[1, 0]
		val d = input[1, 0]
		return a * d - b * c
	} else {
		val cached = cache.load(input.columns - 1, Float2DArray::allocate)
		var result = 0f

		for (i in 0 until input.columns) {
			loadMatrixMinor(input, cached, Float2DArray::get, Float2DArray::set, i, 0)

			if (i and 1 == 0) {
				result += matrixDeterminant(cached, cache)
			} else {
				result -= matrixDeterminant(cached, cache)
			}
		}

		return result
	}
}

internal fun matrixDeterminant(input: Double2DArray, cache: ArrayList<Reference<Double2DArray>?> = doubleCache.get()): Double {
	input.requireIsSquare()
	require(input.columns >= 1) { "Can't compute determinant of matrix with no elements" }

	if (input.columns == 1) {
		return input[0, 0]
	} else if (input.columns == 2) {
		val a = input[0, 0]
		val b = input[0, 1]
		val c = input[1, 0]
		val d = input[1, 0]
		return a * d - b * c
	} else {
		val cached = cache.load(input.columns - 1, Double2DArray::allocate)
		var result = 0.0

		for (i in 0 until input.columns) {
			loadMatrixMinor(input, cached, Double2DArray::get, Double2DArray::set, i, 0)

			if (i and 1 == 0) {
				result += matrixDeterminant(cached, cache)
			} else {
				result -= matrixDeterminant(cached, cache)
			}
		}

		return result
	}
}

internal fun matrixDeterminant(input: Int2DArray, cache: ArrayList<Reference<Int2DArray>?> = intCache.get()): Int {
	input.requireIsSquare()
	require(input.columns >= 1) { "Can't compute determinant of matrix with no elements" }

	if (input.columns == 1) {
		return input[0, 0]
	} else if (input.columns == 2) {
		val a = input[0, 0]
		val b = input[0, 1]
		val c = input[1, 0]
		val d = input[1, 0]
		return a * d - b * c
	} else {
		val cached = cache.load(input.columns - 1, Int2DArray::allocate)
		var result = 0

		for (i in 0 until input.columns) {
			loadMatrixMinor(input, cached, Int2DArray::get, Int2DArray::set, i, 0)

			if (i and 1 == 0) {
				result += matrixDeterminant(cached, cache)
			} else {
				result -= matrixDeterminant(cached, cache)
			}
		}

		return result
	}
}

internal fun matrixDeterminant(input: Long2DArray, cache: ArrayList<Reference<Long2DArray>?> = longCache.get()): Long {
	input.requireIsSquare()
	require(input.columns >= 1) { "Can't compute determinant of matrix with no elements" }

	if (input.columns == 1) {
		return input[0, 0]
	} else if (input.columns == 2) {
		val a = input[0, 0]
		val b = input[0, 1]
		val c = input[1, 0]
		val d = input[1, 0]
		return a * d - b * c
	} else {
		val cached = cache.load(input.columns - 1, Long2DArray::allocate)
		var result = 0L

		for (i in 0 until input.columns) {
			loadMatrixMinor(input, cached, Long2DArray::get, Long2DArray::set, i, 0)

			if (i and 1 == 0) {
				result += matrixDeterminant(cached, cache)
			} else {
				result -= matrixDeterminant(cached, cache)
			}
		}

		return result
	}
}

inline fun <N : Number, M : Array2D> mulMatrixComponents(
	a: M, b: M,
	get: (M, Int, Int) -> N,
	times: (N, N) -> N,
	plus: (N, N) -> N,
	column: Int,
	row: Int
): N {
	require(a.columns == b.rows) { "Can't multiply incompatible matrices: A has ${a.columns} columns, B has ${b.rows} rows" }
	var result = times(get(a, 0, row), get(b, column, 0))

	for (i in 1 until a.columns) {
		result = plus(result, times(get(a, i, row), get(b, column, i)))
	}

	return result
}

fun mulMatrixComponents(a: Float2DArray, b: Float2DArray, column: Int, row: Int): Float {
	return mulMatrixComponents(a, b, Float2DArray::get, Float::times, Float::plus, column, row)
}

fun mulMatrixComponents(a: Double2DArray, b: Double2DArray, column: Int, row: Int): Double {
	return mulMatrixComponents(a, b, Double2DArray::get, Double::times, Double::plus, column, row)
}

internal inline fun <M : Array2D> matrixToString(input: M, get: (M, Int, Int) -> Any?): String {
	if (input.size == 0) {
		return "[]"
	}

	val elements = ArrayList<String>(input.size)
	var maxWidth = 0

	for (row in input.rowIndices) {
		for (column in input.columnIndices) {
			val elem = get(input, column, row).toString()
			elements.add(elem)
			maxWidth = maxWidth.coerceAtLeast(elem.length)
		}
	}

	maxWidth++

	val builder = StringBuilder()
	builder.append("[\n")
	val iterator = elements.iterator()

	for (row in input.rowIndices) {
		builder.append("\t")

		for (column in input.columnIndices) {
			val elem = iterator.next()
			builder.append(" ".repeat(maxWidth - elem.length))
			builder.append(elem)
			builder.append(",")
		}

		builder.append("\n")
	}

	builder.append("]")

	return builder.toString()
}

internal inline fun <reified M : Array2D> matrixEquals(a: M, b: Any?, equals: (M, M, Int, Int) -> Boolean): Boolean {
	if (b == null || a.javaClass !== b.javaClass) return false
	return matrixContentEquals(a, b as M, equals)
}

internal inline fun <reified M : Array2D> matrixContentEquals(a: M, b: M, equals: (M, M, Int, Int) -> Boolean): Boolean {
	if (a.columns != b.columns || a.rows != b.rows) return false

	for (column in a.columnIndices) {
		for (row in a.rowIndices) {
			if (!equals(a, b, column, row)) {
				return false
			}
		}
	}

	return true
}

internal inline fun <M : Array2D> matrixHashCode(input: M, get: (M, Int, Int) -> Int): Int {
	var result = 0

	for (column in input.columnIndices) {
		for (row in input.rowIndices) {
			result = result * 31 + get(input, column, row)
		}
	}

	return result
}

internal inline fun <M : Array2D, B : Buffer, N> matrixStoreColumnRow(input: M, target: B, get: (M, Int, Int) -> N, set: (B, N) -> Unit) {
	for (column in 0 until input.columns) {
		for (row in 0 until input.rows) {
			set(target, get(input, column, row))
		}
	}
}

internal inline fun <M : Array2D, B : Buffer, N> matrixStoreRowColumn(input: M, target: B, get: (M, Int, Int) -> N, set: (B, N) -> Unit) {
	for (row in 0 until input.rows) {
		for (column in 0 until input.columns) {
			set(target, get(input, column, row))
		}
	}
}

internal inline fun <M : Array2D, B : Buffer, N> matrixLoadColumnRow(input: M, target: B, set: (M, Int, Int, N) -> Unit, get: (B) -> N) {
	for (column in 0 until input.columns) {
		for (row in 0 until input.rows) {
			set(input, column, row, get(target))
		}
	}
}

internal inline fun <M : Array2D, B : Buffer, N> matrixLoadRowColumn(input: M, target: B, set: (M, Int, Int, N) -> Unit, get: (B) -> N) {
	for (row in 0 until input.rows) {
		for (column in 0 until input.columns) {
			set(input, column, row, get(target))
		}
	}
}

internal inline fun <I : Array2D, M : I, N> matrixLoad(from: I, into: M, get: (I, Int, Int) -> N, set: (M, Int, Int, N) -> Unit): M {
	from.requireSizeEquals(into)

	for (column in 0 until from.columns) {
		for (row in 0 until from.rows) {
			set(into, column, row, get(from, column, row))
		}
	}

	return into
}

internal inline fun <I : Array2D, M : I, N> matrixLoadTransposed(from: I, into: M, get: (I, Int, Int) -> N, set: (M, Int, Int, N) -> Unit): M {
	require(from.columns == into.rows && from.rows == into.columns) { "Incompatible matrix for transposed load: from: ${from.columns} x ${from.rows}; into: ${into.columns} x ${into.rows}" }

	for (column in 0 until from.columns) {
		for (row in 0 until from.rows) {
			set(into, row, column, get(from, column, row))
		}
	}

	return into
}

internal inline fun <I : Array2D, M : I, N : Number> matrixPlainOperator(
	a: I, b: I,
	get: (I, Int, Int) -> N,
	operator: (N, N) -> N,
	set: (M, Int, Int, N) -> Unit,
	factory: (Int, Int) -> M
): M {
	a.requireSizeEquals(b)
	val new = factory(a.columns, a.rows)

	for (column in 0 until a.columns) {
		for (row in 0 until a.rows) {
			set(new, column, row, operator(get(a, column, row), get(b, column, row)))
		}
	}

	return new
}

internal inline fun <I : Array2D, M : I, N : Number> matrixPlainOperator(
	a: I, b: N,
	get: (I, Int, Int) -> N,
	operator: (N, N) -> N,
	set: (M, Int, Int, N) -> Unit,
	factory: (Int, Int) -> M
): M {
	val new = factory(a.columns, a.rows)

	for (column in 0 until a.columns) {
		for (row in 0 until a.rows) {
			set(new, column, row, operator(get(a, column, row), b))
		}
	}

	return new
}

internal inline fun <I : Array2D, M : I, N : Number> matrixPlainOperator(
	a: M, b: I,
	get: (I, Int, Int) -> N,
	operator: (N, N) -> N,
	set: (M, Int, Int, N) -> Unit,
): M {
	a.requireSizeEquals(b)

	for (column in 0 until a.columns) {
		for (row in 0 until a.rows) {
			set(a, column, row, operator(get(a, column, row), get(b, column, row)))
		}
	}

	return a
}

internal inline fun <M : Array2D, N : Number> matrixPlainOperator(
	a: M, b: N,
	get: (M, Int, Int) -> N,
	operator: (N, N) -> N,
	set: (M, Int, Int, N) -> Unit
): M {
	for (column in 0 until a.columns) {
		for (row in 0 until a.rows) {
			set(a, column, row, operator(get(a, column, row), b))
		}
	}

	return a
}

internal inline fun <I : Array2D, M : I, N : Number> matrixTimes(
	a: I, b: I,
	get: (I, Int, Int) -> N,
	set: (M, Int, Int, N) -> Unit,
	plus: (N, N) -> N,
	times: (N, N) -> N,
	factory: (Int, Int) -> M
): M {
	require(a.columns == b.rows) { "Can't multiply incompatible matrices: A has ${a.columns} columns, B has ${b.rows} rows" }
	val new = factory(b.columns, a.rows)

	for (column in 0 until new.columns) {
		for (row in 0 until new.rows) {
			set(new, column, row, mulMatrixComponents(a, b, get, times, plus, column, row))
		}
	}

	return new
}

private inline fun <N : Number, F : Array2D> getOrElse(
	get: (F, Int, Int) -> N,
	input: F,
	column: Int,
	row: Int,
	orElse: N
): N {
	if (column < input.columns && row < input.rows) {
		return get(input, column, row)
	} else {
		return orElse
	}
}
