package ru.dbotthepony.kommons.io

import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import ru.dbotthepony.kommons.util.KOptional
import java.io.DataInput
import java.io.EOFException
import java.io.IOException
import java.io.InputStream
import java.io.RandomAccessFile
import java.math.BigDecimal
import java.math.BigInteger
import java.util.*
import java.util.function.IntSupplier
import kotlin.collections.ArrayList

fun InputStream.readBigDecimal(): BigDecimal {
	val scale = readInt()
	val size = readVarInt()
	require(size >= 0) { "Negative payload size: $size" }
	val bytes = ByteArray(size)
	read(bytes)
	return BigDecimal(BigInteger(bytes), scale)
}

fun <S : InputStream, V, C : MutableCollection<V>> S.readCollection(reader: S.() -> V, factory: (Int) -> C): C {
	val size = readVarInt()
	val collection = factory.invoke(size)

	for (i in 0 until size) {
		collection.add(reader(this))
	}

	return collection
}

fun <S : InputStream, V> S.readCollection(reader: S.() -> V) = readCollection(reader, ::ArrayList)

private fun InputStream.readOrError(): Int {
	val read = read()

	if (read == -1)
		throw EOFException()

	return read
}

private fun InputStream.readOrError(bytes: ByteArray): ByteArray {
	val read = read(bytes)
	check(read == bytes.size) { "Expected to read ${bytes.size} bytes, but only $read were read" }
	return bytes
}

fun InputStream.readInt(): Int {
	if (this is DataInput) {
		return readInt()
	}

	return (readOrError() shl 24) or (readOrError() shl 16) or (readOrError() shl 8) or readOrError()
}

fun InputStream.readLong(): Long {
	if (this is DataInput) {
		return readLong()
	}

	return (readOrError().toLong() shl 48) or
		(readOrError().toLong() shl 40) or
		(readOrError().toLong() shl 32) or
		(readOrError().toLong() shl 24) or
		(readOrError().toLong() shl 16) or
		(readOrError().toLong() shl 8) or
		readOrError().toLong()
}

fun InputStream.readFloat() = Float.fromBits(readInt())
fun InputStream.readDouble() = Double.fromBits(readLong())

data class VarIntReadResult(val value: Int, val cells: Int)
data class VarLongReadResult(val value: Long, val cells: Int)

private inline fun readVarLongInfo(supplier: () -> Int): VarLongReadResult {
	var result = 0L
	var read = supplier()
	var i = 1

	if (read == -1)
		throw EOFException()

	while (true) {
		result = (result shl 7) or (read.toLong() and 0x7FL)

		if (read and 0x80 == 0)
			break

		read = supplier()

		if (read == -1)
			throw EOFException("Reached end of stream before finished reading variable length integer (read totally $i bytes)")

		i++
	}

	return VarLongReadResult(result, i)
}

private inline fun readVarIntInfo(supplier: () -> Int): VarIntReadResult {
	val read = readVarLongInfo(supplier)
	return VarIntReadResult(read.value.toInt(), read.cells)
}

private inline fun readVarInt(supplier: () -> Int): Int {
	return readVarIntInfo(supplier).value
}

private inline fun readVarLong(supplier: () -> Int): Long {
	return readVarLongInfo(supplier).value
}

private fun Int.fromSignedVar(): Int {
	val sign = this and 1

	if (sign == 0)
		return this ushr 1
	else
		return -(this ushr 1) - 1
}

private fun VarIntReadResult.fromSignedVar(): VarIntReadResult {
	val sign = this.value and 1

	if (sign == 0)
		return VarIntReadResult(this.value ushr 1, this.cells)
	else
		return VarIntReadResult(-(this.value ushr 1) - 1, this.cells)
}

private fun Long.fromSignedVar(): Long {
	val sign = this and 1

	if (sign == 0L)
		return this ushr 1
	else
		return -(this ushr 1) - 1L
}

private fun VarLongReadResult.fromSignedVar(): VarLongReadResult {
	val sign = this.value and 1

	if (sign == 0L)
		return VarLongReadResult(this.value ushr 1, this.cells)
	else
		return VarLongReadResult(-(this.value ushr 1) - 1, this.cells)
}

/**
 * Reads unsigned variable length integer, with Big-endian byte order
 */
fun RandomAccessFile.readVarLong() = readVarLong(::readUnsignedByte)

/**
 * Reads unsigned variable length integer, with Big-endian byte order
 */
fun RandomAccessFile.readVarInt() = readVarInt(::readUnsignedByte)

/**
 * Reads unsigned variable length integer, with Big-endian byte order
 */
fun RandomAccessFile.readVarLongInfo() = readVarLongInfo(::readUnsignedByte)

/**
 * Reads unsigned variable length integer, with Big-endian byte order
 */
fun RandomAccessFile.readVarIntInfo() = readVarIntInfo(::readUnsignedByte)

/**
 * Reads unsigned variable length integer, with Big-endian byte order
 */
fun InputStream.readVarLong() = readVarLong(::read)

/**
 * Reads unsigned variable length integer, with Big-endian byte order
 */
fun InputStream.readVarInt() = readVarInt(::read)

/**
 * Reads unsigned variable length integer, with Big-endian byte order
 */
fun InputStream.readVarLongInfo() = readVarLongInfo(::read)

/**
 * Reads unsigned variable length integer, with Big-endian byte order
 */
fun InputStream.readVarIntInfo() = readVarIntInfo(::read)


/**
 * Reads signed variable length integer, with Big-endian byte order. Sign is determined by first bit.
 */
fun RandomAccessFile.readSignedVarLong() = readVarLong(::readUnsignedByte).fromSignedVar()

/**
 * Reads signed variable length integer, with Big-endian byte order. Sign is determined by first bit.
 */
fun RandomAccessFile.readSignedVarInt() = readVarInt(::readUnsignedByte).fromSignedVar()

/**
 * Reads signed variable length integer, with Big-endian byte order. Sign is determined by first bit.
 */
fun RandomAccessFile.readSignedVarLongInfo() = readVarLongInfo(::readUnsignedByte).fromSignedVar()

/**
 * Reads signed variable length integer, with Big-endian byte order. Sign is determined by first bit.
 */
fun RandomAccessFile.readSignedVarIntInfo() = readVarIntInfo(::readUnsignedByte).fromSignedVar()

/**
 * Reads signed variable length integer, with Big-endian byte order. Sign is determined by first bit.
 */
fun InputStream.readSignedVarLong() = readVarLong(::read).fromSignedVar()

/**
 * Reads signed variable length integer, with Big-endian byte order. Sign is determined by first bit.
 */
fun InputStream.readSignedVarInt() = readVarInt(::read).fromSignedVar()

/**
 * Reads signed variable length integer, with Big-endian byte order. Sign is determined by first bit.
 */
fun InputStream.readSignedVarLongInfo() = readVarLongInfo(::read).fromSignedVar()

/**
 * Reads signed variable length integer, with Big-endian byte order. Sign is determined by first bit.
 */
fun InputStream.readSignedVarIntInfo() = readVarIntInfo(::read).fromSignedVar()

fun InputStream.readBinaryString(): String {
	val size = readVarInt()
	require(size >= 0) { "Negative payload size: $size" }
	val bytes = ByteArray(size)
	check(read(bytes) == size) { "Reached end of stream" }
	return bytes.decodeToString()
}

fun RandomAccessFile.readString(length: Int): String {
	require(length >= 0) { "Invalid length $length" }

	val bytes = ByteArray(length)

	try {
		readFully(bytes)
	} catch(err: Throwable) {
		throw IOException("Tried to read string with length of $length", err)
	}

	return bytes.toString(Charsets.UTF_8)
}

fun InputStream.readString(length: Int, allowSmaller: Boolean = false): String {
	require(length >= 0) { "Invalid length $length" }

	val bytes = ByteArray(length)

	try {
		val read = read(bytes)

		if (!allowSmaller)
			require(read == bytes.size) { "Read $read bytes, expected ${bytes.size}" }
	} catch(err: Throwable) {
		throw IOException("Tried to read string with length of $length", err)
	}

	return bytes.toString(Charsets.UTF_8)
}

fun InputStream.readUTF(): String {
	val bytes = ByteArrayList()
	var read = read()

	while (read != 0) {
		bytes.add(read.toByte())
		read = read()
	}

	return String(bytes.toByteArray())
}

fun InputStream.readUUID(): UUID {
	return UUID(readLong(), readLong())
}

fun <S : InputStream, M : MutableMap<K, V>, K, V> S.readMap(keyReader: S.() -> K, valueReader: S.() -> V, constructor: (Int) -> M): M {
	val size = readVarInt()
	require(size >= 0) { "Negative payload size: $size" }
	val map = constructor(size)

	for (i in 0 until size) {
		map[keyReader(this)] = valueReader(this)
	}

	return map
}

fun <S : InputStream, K, V> S.readMap(keyReader: S.() -> K, valueReader: S.() -> V): Object2ObjectOpenHashMap<K, V> {
	return readMap(keyReader, valueReader, ::Object2ObjectOpenHashMap)
}

fun <S : InputStream, V> S.readKOptional(reader: S.() -> V): KOptional<V> {
	if (read() == 0) {
		return KOptional.empty()
	} else {
		return KOptional(reader(this))
	}
}

fun <S : InputStream, V : Any> S.readOptional(reader: S.() -> V): Optional<V> {
	if (read() == 0) {
		return Optional.empty()
	} else {
		return Optional.of(reader(this))
	}
}

fun InputStream.readByteArray(maxRead: Int = Int.MAX_VALUE): ByteArray {
	val size = readVarInt()
	if (size > maxRead)
		throw IllegalArgumentException("Trying to read $size bytes when $maxRead is the max allowed")
	return readOrError(ByteArray(size))
}
